import ipaddress
from datetime import datetime

from dns_messages import DnsMessage, DnsServer, Question, OPCODE, RRClass, A, TXT, RCODE, RRType, AAAA, PTR, ResourceRecord, DnsPacketParsingException, CNAME, MX
from abc import abstractmethod
from cli_formatter.output_formatting import info, warning, error

from fakedns.config.configuration import Configuration
from fakedns.config.domain_pattern import DomainPattern
import socket
from typing import Tuple


class FakeDnsServer(DnsServer):
    def __init__(self, configuration: Configuration, log_query_only_once: bool = False, log_domain_only_once: bool = False,
                 nxdomain_response: bool = False, no_response: bool = False, verbose: bool = False, proxy_queries: bool = False):
        server_ip = configuration.get_dns_server_ip()
        server_port = configuration.get_dns_server_port()
        super().__init__(ip_address=server_ip, port=server_port)
        info(message='fakedns starts listening on {}:{}'.format(server_ip, server_port))

        self._configuration = configuration

        self._proxy_queries = proxy_queries
        self._log_query_only_once = log_query_only_once
        self._log_domain_only_once = log_domain_only_once
        self._nxdomain_response = nxdomain_response
        self._no_response = no_response
        self._verbose = verbose

        self._logged_queries = set()

    @abstractmethod
    def _handle_received_message(self, message: DnsMessage, remote_ip: str, remote_port: int) -> None:
        if self._verbose:
            FakeDnsServer._log_incoming_message_verbose(message=message, remote_ip=remote_ip, remote_port=remote_port)

        if not message.is_query():
            return

        if len(message.questions) == 0:
            if self._verbose:
                info(message='received a DNS query from {} without a question'.format(remote_ip))

        question: Question = message.questions[0]

        # check if message is already ignored by a existing pattern
        for domain_pattern in self._configuration.get_domain_patterns():
            if domain_pattern.does_pattern_match(question=question):
                self._answer_query_with_pattern(message=message, question=question, pattern=domain_pattern, remote_ip=remote_ip, remote_port=remote_port)
                return

    @staticmethod
    def _log_incoming_message_verbose(message: DnsMessage, remote_ip: str, remote_port: int) -> None:
        print('\treceived a DNS message from {}:{}'.format(remote_ip, remote_port))

        if message.is_query():
            query_str = 'yes'
        else:
            query_str = 'no '
        print('\tMessageID: {}   -   Query: {}   -   OPCODE: {}'.format(str(message.message_id).ljust(5, ' '), query_str, message.op_code.name))

        if len(message.questions) > 0:
            print('\t\tquestions:')
            question: Question
            for question in message.questions:
                print('\t\t\t - {} - {} - {}'.format(question.rr_class.name.ljust(4, ' '), question.rr_type.name.ljust(4, ' '), question.name))

        def _log_rr(rr: ResourceRecord):
            print('\t\t\t - {} - {} - {} - TTL: {}'.format(rr.rr_class.name.ljust(4, ' '), rr.get_RR_type().name.ljust(4, ' '), rr.name, rr.ttl))
            if isinstance(rr, A) or isinstance(rr, AAAA):
                print('\t\t\t   IP: {}'.format(rr.ip_address))
            if isinstance(rr, TXT):
                for i, line in enumerate(rr.text_lines):
                    if i == 0:
                        print('\t\t\t   TXT: {}'.format(line))
                    else:
                        print('\t\t\t        {}'.format(line))

        if len(message.answers_RRs) > 0:
            print('\t\tanswers:')
            resource_record: ResourceRecord
            for resource_record in message.answers_RRs:
                _log_rr(rr=resource_record)

        if len(message.authority_RRs) > 0:
            print('\t\tauthority RRs:')
            resource_record: ResourceRecord
            for resource_record in message.authority_RRs:
                _log_rr(rr=resource_record)

        if len(message.additional_RRs) > 0:
            print('\t\tadditional RRs:')
            resource_record: ResourceRecord
            for resource_record in message.additional_RRs:
                _log_rr(rr=resource_record)

        print(30 * '-')

    def _answer_query_with_pattern(self, message: DnsMessage, question: Question, pattern: DomainPattern, remote_ip: str, remote_port: int) -> DnsMessage or None:
        answer_message = self._generate_answer_from_pattern(message=message, question=question, pattern=pattern)
        if answer_message is not None:
            self.send_message(message=answer_message, remote_ip=remote_ip, remote_port=remote_port)
        self._log_query(answer_message=answer_message, question=question, pattern=pattern, remote_ip=remote_ip, remote_port=remote_port)
        return answer_message

    def _log_query(self, answer_message: DnsMessage, question: Question, pattern: DomainPattern, remote_ip: str, remote_port: int) -> None:
        if not pattern.log_request:
            return

        if self._log_query_only_once:
            question_as_string: str = question.__str__()
            if question_as_string in self._logged_queries:
                return
            self._logged_queries.add(question_as_string)

        if self._log_domain_only_once:
            if question.name in self._logged_queries:
                return
            self._logged_queries.add(question.name)

        log_line = self._configuration.get_log_format()
        log_line = log_line.replace('%DATETIME%', datetime.now().strftime('%Y-%m-%d %H:%M:%S'))
        log_line = log_line.replace('%RR_TYPE%', question.rr_type.name.ljust(4, ' '))
        log_line = log_line.replace('%RR_CLASS%', question.rr_class.name.ljust(4, ' '))
        log_line = log_line.replace('%DOMAINNAME%', question.name.ljust(35, ' '))
        log_line = log_line.replace('%REMOTE_IP%', remote_ip.ljust(15, ' '))
        log_line = log_line.replace('%REMOTE_PORT%', str(remote_port).rjust(5, ' '))

        proxy_text = ' (from Proxy)' if pattern.proxy_queries or self._proxy_queries else ''

        if answer_message is None:
            log_line = log_line.replace('%RESPONSE%', 'No response')
        elif answer_message.rcode == RCODE.non_existing_domain:
            log_line = log_line.replace('%RESPONSE%', 'Response{}: NXDOMAIN'.format(proxy_text))
        elif len(answer_message.answers_RRs) > 0:
            answer: ResourceRecord = answer_message.answers_RRs[0]
            log_line_response = self._configuration.get_log_format_response()
            log_line_response = log_line_response.replace('%TTL%', '{}'.format(answer.ttl))
            if isinstance(answer, A) or isinstance(answer, AAAA):
                log_line_response = log_line_response.replace('%RR_VALUE%', str(answer.ip_address))
            elif isinstance(answer, TXT):
                log_line_response = log_line_response.replace('%RR_VALUE%', '{}'.format(answer.text_lines))
            elif isinstance(answer, PTR):
                log_line_response = log_line_response.replace('%RR_VALUE%', '{}'.format(answer.domain_name))
            elif isinstance(answer, CNAME):
                log_line_response = log_line_response.replace('%RR_VALUE%', 'CNAME: {}'.format(answer.domain_name))
            elif isinstance(answer, MX):
                log_line_response = log_line_response.replace('%RR_VALUE%', 'MX: {}'.format(answer.exchange))

            log_line = log_line.replace('%RESPONSE%', 'Response{}: '.format(proxy_text) + log_line_response)
        else:
            log_line = log_line.replace('%RESPONSE%', 'Response{}: No Record'.format(proxy_text))

        info(log_line)

    def _generate_answer_from_pattern(self, message: DnsMessage, question: Question, pattern: DomainPattern) -> DnsMessage or None:
        answer = DnsMessage(message_id=message.message_id, qr=1)
        answer.questions.append(question)

        if pattern.not_existing_domain or self._nxdomain_response:
            answer.rcode = RCODE.non_existing_domain
            return answer

        if self._no_response:
            return None

        if pattern.proxy_queries or self._proxy_queries:
            return self._proxy_dns_query(message=message)

        rrtype: RRType = RRType[question.rr_type.name]
        answer_value = pattern.answer_dict.get(rrtype.name, None)
        if answer_value is None:
            return answer

        if rrtype == RRType.A:
            rr = A(name=question.name, rr_class=RRClass.IN, ip_address=ipaddress.IPv4Address(answer_value), ttl=pattern.ttl)
        elif rrtype == RRType.AAAA:
            rr = AAAA(name=question.name, rr_class=RRClass.IN, ip_address=ipaddress.IPv6Address(answer_value), ttl=pattern.ttl)
        elif rrtype == RRType.PTR:
            rr = PTR(name=question.name, rr_class=RRClass.IN, domain_name=answer_value, ttl=pattern.ttl)
        elif rrtype == RRType.TXT:
            rr = TXT(name=question.name, rr_class=RRClass.IN, text_lines=answer_value, ttl=pattern.ttl)
        else:
            warning('Answer to RR Type "{}" is currently not supported'.format(rrtype.name))
            return None

        answer.answers_RRs.append(rr)
        return answer

    def _proxy_dns_query(self, message: DnsMessage) -> DnsMessage or None:
        proxy_address = self._configuration.get_dns_proxy_address()

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.settimeout(5)
        try:
            sock.sendto(message.to_bytes(), proxy_address)
        except Exception as exception:
            error(message='message could not be forwarded to dns server {}:{} - error: {}'.format(proxy_address[0], proxy_address[1], exception))
            return

        try:
            raw_bytes, _ = sock.recvfrom(1024)
        except Exception as exception:
            error(message='error while receiving answer from proxy server: {}'.format(exception))
            return

        try:
            return DnsMessage.from_bytes(raw_bytes=raw_bytes)
        except DnsPacketParsingException as exception:
            error(message='error while parsing dns answer from proxy server: {}'.format(exception))
            return





import ipaddress

from dns_messages import RRType

from fakedns.config import DomainPattern
from cli_formatter.output_formatting import warning, error, info
from typing import List


def add_pattern_interactively() -> DomainPattern:
    name_pattern = input('name pattern > ').strip()
    if len(name_pattern) == 0:
        error(message='pattern can not be empty')

    priority = _cli_input_integer(text='priority', default_value=10)
    type_filter = _cli_input_type_filter()
    log_request = _cli_input_boolean(text='log request', default_value=True)
    proxy_queries = _cli_input_boolean(text='proxy queries', default_value=False)
    not_existing_domain = _cli_input_boolean(text='respond with nxdomain error', default_value=False)
    ttl = _cli_input_integer(text='TTL in seconds', default_value=3600)

    answer_dict = {
        'A': _cli_input_ipv4_address(),
        'AAAA': _cli_input_ipv6_address(),
    }

    answer_PTR = _cli_input_PTR()
    if answer_PTR is not None:
        answer_dict['PTR'] = answer_PTR

    answer_TXT = _cli_input_TXT()
    if answer_TXT is not None:
        answer_dict['TXT'] = answer_TXT

    return DomainPattern(priority=priority, name_pattern=name_pattern, type_filter=type_filter, proxy_queries=proxy_queries,
                         not_existing_domain=not_existing_domain, log_request=log_request, answer_dict=answer_dict, ttl=ttl)


def _cli_input_type_filter() -> List[RRType]:
    value = input('apply pattern only to queries regarding specific RR types (comma separated list) [default=None] > ').strip()
    if len(value) == 0:
        return list()

    type_filter = list()
    for part in value.split(','):
        part = part.strip()
        try:
            type_filter.append(RRType[part])
        except AttributeError:
            error(message='RR type "{}" does not exist'.format(part))
            exit()
    return type_filter


def _cli_input_ipv4_address() -> str:
    value = input('value of A record [DefaultIPv4 / IPv4 Address] > ').strip().lower()
    if len(value) == 0 or value == 'defaultipv4':
        return 'DefaultIPv4'
    try:
        ipaddress.IPv4Address(value)
    except AttributeError:
        error('input is not a valid IPv4 address')
        exit()
    return value


def _cli_input_ipv6_address() -> str or None:
    value = input('value of A record [DefaultIPv6 / IPv6 Address] > ').strip().lower()
    if len(value) == 0 or value == 'defaultipv6':
        return 'DefaultIPv6'
    try:
        ipaddress.IPv6Address(value)
    except AttributeError:
        error('input is not a valid IPv6 address')
        exit()
    return value


def _cli_input_PTR() -> str or None:
    value = input('value of PTR record [default=None] > ').strip().lower()
    if len(value) == 0 or value == 'none':
        return None
    return value


def _cli_input_TXT() -> List[bytes] or None:
    value = input('value of TXT record (lines separated with ";") [default=None] > ').strip().lower()
    if len(value) == 0 or value == 'none':
        return None
    return [x.encode('ascii') for x in value.split(';')]


def _cli_input_integer(text: str, default_value: int) -> int:
    default_str = ' [default={}]'.format(default_value)
    value = input('{}{} > '.format(text, default_str)).strip()
    if len(value) == 0:
        return default_value
    try:
        return int(value)
    except AttributeError:
        error(message="input has to be a integer")
        exit()


def _cli_input_boolean(text: str, default_value: bool) -> bool:
    default_str = ' [Y/n]' if default_value else ' [N/y]'
    value = input('{}{} > '.format(text, default_str)).strip().lower()
    if len(value) == 0:
        return default_value
    if value == 'y':
        return True
    if value == 'n':
        return False
    error(message='input is not valid - only "y" or "n" are accepted')
    exit()




import json
from sys import platform
import shutil
import os
import socket
import ipaddress
from typing import List, Iterator, Tuple
from cli_formatter.table_builder import TableBuilderClassic
from cli_formatter.output_formatting import colorize_string, Color, info
from dns_messages import Question, RRType

from .add_pattern_interactively import add_pattern_interactively
from .domain_pattern import DomainPattern


class OperatingSystemNotDetected(Exception):
    def __init__(self, platform_name: str):
        self.platform_name = platform_name


class DefaultConfigFileNotFound(Exception):
    def __init__(self, path: str):
        self.path = path


class FileAlreadyExists(Exception):
    def __init__(self, path: str):
        self.path = path


class ConfigLoadError(Exception):
    def __init__(self, message: str):
        self.message = message


class InvalidConfig(Exception):
    def __init__(self, message: str):
        self.message = message


class Configuration:
    def __init__(self):
        self._path_to_configuration_file: str or None = None

        self._default_ipv4 = None
        self._default_ipv6 = None
        self._server_port = 53
        self._server_ip = None
        self._log_format = '%DATETIME% - %RR_TYPE% - %DOMAIN_NAME% => %RESPONSE%'
        self._log_format_of_response = 'TTL: %TTL% - %RR_VALUE%'
        self._domain_patterns: List[DomainPattern] = list()
        self._dns_proxy_address = None

    @staticmethod
    def _determine_path_to_config_file(user_input_path_to_configuration_file: str or None, create_file: bool = False) -> str:
        # if the user specified a specific config file to load then load this file
        if user_input_path_to_configuration_file is not None:
            return user_input_path_to_configuration_file

        # if the user has no config file specified then user the global one
        if platform == "linux" or platform == "linux2" or platform == "darwin":
            path_to_config_folder = '/etc/fakedns'
        elif platform == "win32":
            path_to_config_folder = os.getenv('ProgramData') + os.sep + 'fakedns'
        else:
            raise OperatingSystemNotDetected(platform_name=platform)

        path_to_file = path_to_config_folder + os.sep + 'global.conf'

        if not os.path.exists(path_to_config_folder):
            if create_file:
                os.mkdir(path_to_config_folder)
            else:
                raise DefaultConfigFileNotFound(path=path_to_file)
        elif not os.path.isdir(path_to_config_folder):
            raise ConfigLoadError('"{}" is not a directory, thus the config which is expected to be in this directory could not be loaded'.format(path_to_config_folder))

        if not os.path.exists(path_to_file) and not create_file:
            raise DefaultConfigFileNotFound(path=path_to_file)
        elif os.path.exists(path_to_file) and create_file:
            raise FileAlreadyExists(path=path_to_file)
        elif os.path.isdir(path_to_file):
            raise ConfigLoadError('"{}" is a directory and not a config file, thus the config could not be loaded'.format(path_to_file))

        return path_to_file

    def load(self, user_input_path_to_configuration_file: str or None = None) -> None:
        self._load_config_file(user_input_path_to_configuration_file=user_input_path_to_configuration_file)

    def _load_config_file(self, user_input_path_to_configuration_file: str) -> None:
        self._path_to_configuration_file: str = Configuration._determine_path_to_config_file(user_input_path_to_configuration_file=user_input_path_to_configuration_file)

        section_list = list()
        current_section = dict()
        section_name = None
        with open(self._path_to_configuration_file, mode='r', encoding='utf-8') as input_file:
            for line_number, line in enumerate(input_file.readlines()):
                line = line.strip()

                # ignore empty lines and comment lines
                if len(line) == 0 or line[0] == '#':
                    continue

                # start of a new section
                if line[0] == '[':
                    # store old section
                    if len(current_section) > 0:
                        section_list.append({'name': section_name, 'data': current_section})
                    current_section = dict()
                    section_name = line[1:-1]
                else:
                    if '=' not in line:
                        raise InvalidConfig(message='Missing "=" in line {}:  "{}"'.format(line_number, line))

                    (key, value) = line.split('=', 1)
                    key = key.strip()
                    value = value.strip()
                    current_section[key] = value

            if len(current_section) > 0:
                if section_name is None:
                    raise InvalidConfig(message='There are values without a specified section')
                section_list.append({'name': section_name, 'data': current_section})

        self._process_config_section_list(section_list=section_list)

    def _process_config_section_list(self, section_list: List[dict]) -> None:
        # the Settings section must be parsed at first since the pattern for example accesses attributes parsed in Settings sections
        self._process_config_settings_section(section_list=section_list)
        self._process_config_output_section(section_list=section_list)
        self._process_config_pattern_entries(section_list=section_list)

    @staticmethod
    def _get_section_by_name(section_name: str, section_list: List[dict]) -> dict:
        for section in section_list:
            if section['name'] == section_name:
                return section['data']
        return dict()

    @staticmethod
    def _get_sections_by_name(section_name: str, section_list: List[dict]) -> Iterator[dict]:
        section: dict
        for section in section_list:
            if section['name'] == section_name:
                yield section['data']

    def _process_config_settings_section(self, section_list: List[dict]) -> None:
        section_dict = Configuration._get_section_by_name(section_name='Settings', section_list=section_list)

        ip_of_this_machine = Configuration._ip_of_this_machine()

        # Load Default IPv4 Address
        if 'DefaultIPv4' in section_dict:
            self._default_ipv4 = section_dict['DefaultIPv4']
        else:
            if isinstance(ip_of_this_machine, ipaddress.IPv4Address):
                self._default_ipv4 = str(ip_of_this_machine)

        # Load Default IPv6 Address
        if 'DefaultIPv6' in section_dict:
            self._default_ipv6 = section_dict['DefaultIPv6']
        else:
            if isinstance(ip_of_this_machine, ipaddress.IPv6Address):
                self._default_ipv6 = str(ip_of_this_machine)

        if 'ListenOnPort' in section_dict:
            self._server_port = int(section_dict['ListenOnPort'])

        if 'ListenOnIP' in section_dict:
            self._server_ip = section_dict['ListenOnIP']
        else:
            if ip_of_this_machine is None:
                raise ConfigLoadError(message='Could not load the IP of this machine, please specify the IP of the fake dns server in the config file')
            else:
                self._server_ip = str(ip_of_this_machine)

        if 'DnsProxyIP' in section_dict:
            dns_proxy_address_ip = section_dict['DnsProxyIP']
        else:
            dns_proxy_address_ip = '8.8.8.8'

        if 'DnsProxyPort' in section_dict:
            dns_proxy_address_port = int(section_dict['DnsProxyPort'])
        else:
            dns_proxy_address_port = 53

        self._dns_proxy_address = (dns_proxy_address_ip, dns_proxy_address_port)

    @staticmethod
    def _ip_of_this_machine() -> ipaddress.IPv4Address or ipaddress.IPv6Address or None:
        try:
            hostname = socket.gethostname()
            return ipaddress.ip_address(socket.gethostbyname(hostname))
        except Exception:
            return None

    def _process_config_output_section(self, section_list: List[dict]) -> None:
        section_dict = Configuration._get_section_by_name(section_name='Output', section_list=section_list)
        if 'format' in section_dict:
            self._log_format = section_dict['format']
        if 'response_format' in section_dict:
            self._log_format_of_response = section_dict['response_format']

    def _process_config_pattern_entries(self, section_list: List[dict]) -> None:
        for section in Configuration._get_sections_by_name(section_name='DomainPattern', section_list=section_list):
            pattern_entry = self._process_config_pattern_entry(config=section)
            self._domain_patterns.append(pattern_entry)

        if len(self._domain_patterns) == 0:
            self._domain_patterns.append(self._get_default_domain_pattern())

        self._domain_patterns = sorted(self._domain_patterns, key=lambda x: x.priority)

    def _process_config_pattern_entry(self, config: dict) -> DomainPattern:
        name_pattern = config.get('name_pattern', '*')
        priority = int(config.get('priority', 99999))

        type_filter = list()
        if 'type_filter' in config:
            for rr_type in config['type_filter'].split(','):
                rr_type = rr_type.strip()
                if len(rr_type) > 0:
                    type_filter.append(RRType[rr_type])

        log_request = config.get('log_request', 'yes') == 'yes'
        proxy_queries = config.get('proxy_queries', 'no') == 'yes'
        not_existing_domain = config.get('not_existing_domain', 'no') == 'yes'
        ttl = int(config.get('ttl', 3600))

        answer_dict = dict()

        answer_A = config.get('answer_A', 'DefaultIPv4')
        if answer_A == 'DefaultIPv4':
            answer_dict['A'] = self.get_default_IPv4()
        else:
            answer_dict['A'] = answer_A

        answer_AAAA = config.get('answer_AAAA', 'DefaultIPv6')
        if answer_AAAA == 'DefaultIPv6':
            answer_dict['AAAA'] = self.get_default_IPv6()
        else:
            answer_dict['AAAA'] = answer_AAAA

        answer_PTR = config.get('answer_PTR', None)
        if answer_PTR is not None:
            answer_dict['PTR'] = answer_PTR

        answer_TXT = config.get('answer_TXT', None)
        if answer_TXT is not None:
            text_lines = json.loads(answer_TXT)
            if not isinstance(text_lines, list):
                raise InvalidConfig('value of "answer_TXT" key must be a list of string in json format')
            answer_dict['TXT'] = [x.encode('ascii') for x in text_lines]

        return DomainPattern(priority=priority, name_pattern=name_pattern, type_filter=type_filter, proxy_queries=proxy_queries,
                             not_existing_domain=not_existing_domain, log_request=log_request, answer_dict=answer_dict, ttl=ttl)

    def _get_default_domain_pattern(self) -> DomainPattern:
        answer_dict = {'TXT': ['FAKE_DNS_TXT'.encode('ascii')]}
        if self.get_default_IPv4() is not None:
            answer_dict['A'] = self.get_default_IPv4()
        if self.get_default_IPv6() is not None:
            answer_dict['AAAA'] = self.get_default_IPv6()
        return DomainPattern(priority=9999, name_pattern='*', type_filter=None, log_request=True, proxy_queries=False, not_existing_domain=False, answer_dict=answer_dict, ttl=3600)

    def get_default_IPv4(self) -> str or None:
        return self._default_ipv4

    def get_default_IPv6(self) -> str or None:
        return self._default_ipv6

    def get_log_format(self) -> str:
        return self._log_format

    def get_log_format_response(self) -> str:
        return self._log_format_of_response

    def get_dns_server_ip(self) -> str:
        return self._server_ip

    def get_dns_server_port(self) -> int:
        return self._server_port

    def get_dns_proxy_address(self) -> Tuple[str, int]:
        return self._dns_proxy_address

    def get_domain_patterns(self) -> List[DomainPattern]:
        return self._domain_patterns

    @staticmethod
    def command_init(path: str or None = None):
        path_to_default_config = os.path.dirname(__file__) + os.sep + 'default_config.config'
        path_to_config = Configuration._determine_path_to_config_file(user_input_path_to_configuration_file=path, create_file=True)
        shutil.copyfile(path_to_default_config, path_to_config)

    @staticmethod
    def command_edit(path: str or None = None):
        path = Configuration._determine_path_to_config_file(user_input_path_to_configuration_file=path)
        try:
            if platform == "win32":
                os.system('notepad "{}"'.format(path))
            else:   # linux
                os.system('sensible-editor "{}"'.format(path))
        except KeyboardInterrupt:
            pass

    @staticmethod
    def command_fork(destination_path: str):
        original_path = Configuration._determine_path_to_config_file(user_input_path_to_configuration_file=None)
        shutil.copyfile(original_path, destination_path)

    def command_list_patterns(self):
        entry: DomainPattern
        number_of_entries = 0
        data = list()
        for entry in self.get_domain_patterns():
            number_of_entries += 1

            if entry.log_request:
                log_request = colorize_string(text='yes', color=Color.GREEN)
            else:
                log_request = colorize_string(text='no', color=Color.RED)

            if entry.type_filter is None:
                type_filter = '-'
            else:
                type_filter = colorize_string(text=', '.join(entry.type_filter), color=Color.CYAN)

            if entry.proxy_queries:
                proxy_queries = colorize_string(text='proxy', color=Color.YELLOW)
            else:
                proxy_queries = ''

            if entry.not_existing_domain:
                not_existing_domain = colorize_string(text='nxdomain', color=Color.RED)
            else:
                not_existing_domain = ''

            data.append([str(entry.priority), entry.original_name_pattern, type_filter, log_request, proxy_queries, not_existing_domain, str(entry.ttl)])

        header = ['Priority', 'Pattern', 'RR Filter', 'Logging', 'Queries are Proxied', 'NxDomain', 'TTL']
        classic_builder = TableBuilderClassic()
        classic_builder.build_table(header=header, data=data)
        print('config files contains {} patterns'.format(number_of_entries))

    @staticmethod
    def command_add_pattern(path: str or None = None):
        try:
            config = Configuration()
            config.load(user_input_path_to_configuration_file=path)
            new_pattern = add_pattern_interactively()
            config._add_pattern_to_config(pattern=new_pattern)
        except KeyboardInterrupt:
            pass

    def add_new_baseline_pattern(self, question: Question, default_pattern: DomainPattern or None) -> DomainPattern:
        if default_pattern is None:
            new_domain_pattern = DomainPattern(priority=0, name_pattern=question.name, type_filter=[question.rr_type],
                                               proxy_queries=False, not_existing_domain=False, log_request=False,
                                               ttl=3600, answer_dict={
                    RRType.A: self.get_default_IPv4(),
                    RRType.AAAA: self.get_default_IPv6()
                })
        else:
            new_domain_pattern = DomainPattern(priority=0, name_pattern=question.name, type_filter=[question.rr_type],
                                               proxy_queries=default_pattern.proxy_queries,
                                               not_existing_domain=default_pattern.not_existing_domain, log_request=False,
                                               ttl=default_pattern.ttl, answer_dict=default_pattern.answer_dict)

        self._domain_patterns.append(new_domain_pattern)
        self._domain_patterns = sorted(self._domain_patterns, key=lambda x: x.priority)

        self._add_pattern_to_config(pattern=new_domain_pattern)

        return new_domain_pattern

    def _add_pattern_to_config(self, pattern: DomainPattern):
        with open(self._path_to_configuration_file, mode='a') as config_file:
            config_file.write('\n')
            config_file.write('[DomainPattern]\n')
            config_file.write('priority = {}\n'.format(pattern.priority))
            config_file.write('name_pattern = {}\n'.format(pattern.original_name_pattern))

            if len(pattern.type_filter) > 0:
                config_file.write('type_filter = {}\n'.format(','.join([x.name for x in pattern.type_filter])))

            if pattern.log_request:
                config_file.write('log_request = yes\n')
            else:
                config_file.write('log_request = no\n')

            if pattern.proxy_queries:
                config_file.write('proxy_queries = yes\n')

            if pattern.not_existing_domain:
                config_file.write('not_existing_domain = yes\n')

            if pattern.not_existing_domain:
                config_file.write('ttl = {}\n'.format(pattern.ttl))

            for r in pattern.answer_dict:
                if r == 'A' and pattern.answer_dict[r] != self.get_default_IPv4():
                    config_file.write('answer_A = {}\n'.format(pattern.answer_dict[r]))
                if r == 'AAAA' and pattern.answer_dict[r] != self.get_default_IPv6():
                    config_file.write('answer_AAAA = {}\n'.format(pattern.answer_dict[r]))
                if r == 'TXT':
                    config_file.write('answer_TXT = {}\n'.format(pattern.answer_dict[r]))
                if r == 'PTR':
                    config_file.write('answer_PTR = {}\n'.format(pattern.answer_dict[r]))




import time

from fakedns.config.configuration import Configuration
from cli_formatter.output_formatting import colorize_string, Color, warning, info

from fakedns.library.baseline_creation_server import BaselineCreationServer
from fakedns.docopt import docopt, DocoptExit


description = """fakedns-config

Description:
fakedns-config is a script for easy editing and manipulating config files used
by the fakedns script.

Usage:
  fakedns-config help
                        > displays the help page with detailed command descriptions

  fakedns-config edit [<path_to_config>]
                        > opens a text editor to edit the config file manually

  fakedns-config init [<path_to_config>]
                        > create a new default config

  fakedns-config fork [<path_to_new_config_file>]
                        > copies the global config file and saves it at the
                        > given location. Useful if config has to be temporarily
                        > customized for a special use case

  fakedns-config pattern add [<path_to_config>]
                        > a new pattern is added to the config interactively
                        > so that the config has not to be changed manually

  fakedns-config pattern show [<path_to_config>]
                        > lists all patterns which the config file contains
                        > in a table format which makes it suitable to quickly
                        > check the correctness of the used patterns and their
                        > attributes

  fakedns-config pattern baseline [<path_to_config>]
                        > listens to incoming queries and whitelists all received
                        > query patterns in the selected config file until the
                        > user aborts the process via Strg + C.
                        > This command makes it easy to create a baseline
                        > and filter out the "normal" noise from a machine of interest
                        > (e.g. baseline the DNS-requests made from a Windows machine
                        > in a laboratory environment before executing malware on the
                        > machine - in this way only requests are shown which were
                        > triggered by the malware)


Note:
If no path to config is given the default location of the global config is used, which is
/etc/fakedns/global.conf in Linux or %ProgramData%\\fakedns\\global.conf under Windows
"""

version = """
fakedns-config (from fakedns version 1.1)

developed by Florian Wahl (https://www.linkedin.com/in/florian-wahl-security-expert/)

see for more information  https://github.com/wahlflo/fakedns
"""


def print_usage() -> None:
    colored_lines = list()
    for line in description.split('Usage:')[1].split('Note:')[0].split('\n'):
        if line.startswith('  fakedns-config'):
            colored_lines.append(line)
    print('Usage: \n' + '\n'.join(colored_lines))


def print_help() -> None:
    colored_lines = list()
    for line in description.split('\n', 2)[2].split('\n'):
        if line.startswith('Note:') or line.startswith('Description:') or line.startswith('Usage:'):
            colored_lines.append(colorize_string(text=line, color=Color.CYAN))
        elif line.startswith('  fakedns-config'):
            colored_lines.append(colorize_string(text=line, color=Color.BLUE))
        else:
            colored_lines.append(line)
    print('\n'.join(colored_lines), end='')


def main():
    try:
        arguments: dict = docopt(help=False, doc=description, version=version)
    except DocoptExit:
        print_usage()
        return

    if arguments['help']:
        print_help()
        return

    if arguments['edit']:
        path_to_config = arguments['<path_to_config>'] if isinstance(arguments['<path_to_config>'], str) else None
        Configuration.command_edit(path=path_to_config)
        return

    if arguments['init']:
        path_to_config = arguments['<path_to_config>'] if isinstance(arguments['<path_to_config>'], str) else None
        Configuration.command_init(path=path_to_config)
        return

    if arguments['fork']:
        Configuration.command_fork(destination_path=arguments['<path_to_new_config_file>'])
        return

    if arguments['pattern']:
        if arguments['add']:
            path_to_config = arguments['<path_to_config>'] if isinstance(arguments['<path_to_config>'], str) else None
            config = Configuration()
            config.load(user_input_path_to_configuration_file=path_to_config)
            config.command_add_pattern()
        if arguments['show']:
            path_to_config = arguments['<path_to_config>'] if isinstance(arguments['<path_to_config>'], str) else None
            config = Configuration()
            config.load(user_input_path_to_configuration_file=path_to_config)
            config.command_list_patterns()
        if arguments['baseline']:
            path_to_config = arguments['<path_to_config>'] if isinstance(arguments['<path_to_config>'], str) else None
            config = Configuration()
            config.load(user_input_path_to_configuration_file=path_to_config)

            info('Set the values of the new baseline patterns:')

            value_str = input("answer incoming queries using the default pattern [Y/n] ").strip()
            if value_str == 'Y' or value_str == 'y' or value_str == '':
                answer_queries_value = True
            elif value_str == 'N' or value_str == 'n':
                answer_queries_value = False
            else:
                warning('Abort')
                return

            server = BaselineCreationServer(configuration=config, answer_queries=answer_queries_value)
            server.start()
            while True:
                try:
                    time.sleep(500)
                except KeyboardInterrupt:
                    print()
                    warning('received keyboard interrupt')
                    info('DNS server will be stopped...')
                    server.stop()
                    info('DNS server successfully stopped')
                    info('{} baseline patterns were added to the config file'.format(server.get_number_of_added_patterns()))
                    break



from cli_formatter.output_formatting import warning, info, error

from fakedns.config import (
    OperatingSystemNotDetected,
    DefaultConfigFileNotFound,
    FileAlreadyExists,
    ConfigLoadError,
    InvalidConfig,
    Configuration,
)
from fakedns.library import FakeDnsServer
import time
import argparse

usage = 'fakedns [OPTIONS...]'

description = "fakedns is a script which mimicks DNS resolution"

version = """
fakedns version 1.1

developed by Florian Wahl (https://www.linkedin.com/in/florian-wahl-security-expert/)

see for more information  https://github.com/wahlflo/fakedns
"""


def main():
    argument_parser = argparse.ArgumentParser(usage=usage, description=description)
    argument_parser.add_argument('-c', '--config', dest='config', help="path to config file (if not set the default global one is used)", type=str)
    argument_parser.add_argument('--log-query-only-once', dest='log_query_only_once', action='store_true', default=False, help="prevents that the same query is logged multiple times")
    argument_parser.add_argument('--log-domain-only-once', dest='log_domain_only_once', action='store_true', default=False, help="prevents that the same domain name is logged multiple times")
    argument_parser.add_argument('--nxdomain-response', dest='nxdomain_response', action='store_true', default=False, help="respond to all queries with an nxdomain response (overrides settings from the config)")
    argument_parser.add_argument('--no-response', dest='no_response', action='store_true', default=False, help="do not respond to any queries (overrides settings from the config)")
    argument_parser.add_argument('--proxy', dest='proxy', action='store_true', default=False, help="proxy all incoming queries (overrides settings from the config)")
    argument_parser.add_argument('--verbose', dest='verbose', action='store_true', default=False, help="logs more details of each queries")
    argument_parser.add_argument('--version', dest='version', action='store_true', default=False, help="shows version info")
    arguments = argument_parser.parse_args()

    if arguments.version:
        print(version)
        exit()

    try:
        config = Configuration()
        if arguments.config is not None:
            config.load(user_input_path_to_configuration_file=arguments.config)
        else:
            config.load()
    except InvalidConfig as exception:
        error('invalid configuration file: {}'.format(exception.message))
        return
    except ConfigLoadError as exception:
        error('configuration could not be loaded: {}'.format(exception.message))
        return
    except FileAlreadyExists as exception:
        error('configuration file could not be created since file already exists: {}'.format(exception.path))
        return
    except OperatingSystemNotDetected as exception:
        error('the operating system could not be created: {}'.format(exception.platform_name))
        return
    except DefaultConfigFileNotFound as exception:
        error('the default config file was not found: {}  - try generate it with the "fakedns-config" command'.format(exception.path))
        return

    if arguments.no_response and arguments.nxdomain_response:
        error(message="only one of the options '--nxdomain-response' and '--no-response' can be set. ")

    if arguments.log_query_only_once and arguments.log_domain_only_once:
        error(message="only one of the options '--log-query-only-once' and '--log-domain-only-once' can be set. ")

    server = FakeDnsServer(configuration=config, log_query_only_once=arguments.log_query_only_once, log_domain_only_once=arguments.log_domain_only_once,
                           nxdomain_response=arguments.nxdomain_response, no_response=arguments.no_response, verbose=arguments.verbose,
                           proxy_queries=arguments.proxy)
    server.start()
    while True:
        try:
            time.sleep(500)
        except KeyboardInterrupt:
            print()
            warning('received keyboard interrupt')
            info('DNS server will be stopped...')
            server.stop()
            info('DNS server successfully stopped')
            break



from .domain_pattern import DomainPattern
from .configuration import (
    OperatingSystemNotDetected,
    DefaultConfigFileNotFound,
    FileAlreadyExists,
    ConfigLoadError,
    InvalidConfig,
    Configuration,
)



import re
from typing import List, Set
from dns_messages import DnsMessage, Question, RRType


class DomainPattern:
    def __init__(self, priority: int, name_pattern: str, type_filter: Set[RRType] or None, proxy_queries: bool, not_existing_domain: bool, log_request: bool, ttl: int, answer_dict: dict):
        self.priority = priority
        self.type_filter = type_filter
        self.proxy_queries = proxy_queries
        self.ttl = ttl
        self.not_existing_domain = not_existing_domain
        self.log_request = log_request
        self.answer_dict = answer_dict

        self.original_name_pattern = name_pattern
        name_pattern = name_pattern.replace('.', r'\.')
        name_pattern = name_pattern.replace('*', r'.*')
        self.name_pattern = re.compile(name_pattern)

    def __str__(self) -> str:
        return 'DomainPattern[{}]'.format(self.name_pattern)

    def does_pattern_match(self, question: Question) -> bool:
        if len(self.type_filter) > 0 and question.rr_type not in self.type_filter:
            return False
        return self.name_pattern.match(question.name) is not None




from dns_messages import DnsMessage, DnsServer, Question, OPCODE, RRClass, A, TXT
from abc import abstractmethod
from cli_formatter.output_formatting import info

from fakedns.config.configuration import Configuration
from fakedns.library.fake_dns_server import FakeDnsServer


class BaselineCreationServer(FakeDnsServer):
    def __init__(self, configuration: Configuration, answer_queries: bool):
        super().__init__(configuration)
        info(message='incoming queries will be added to the config file as patterns with deactivated logging')
        self._number_of_added_patterns = 0
        self._answer_queries = answer_queries

    @abstractmethod
    def _handle_received_message(self, message: DnsMessage, remote_ip: str, remote_port: int) -> None:
        if not message.is_query() and len(message.questions) > 0:
            return

        question: Question = message.questions[0]

        # check if message is already ignored by a existing pattern
        for domain_pattern in self._configuration.get_domain_patterns():
            if not domain_pattern.log_request and domain_pattern.does_pattern_match(question=question):
                self._answer_query_with_pattern(message=message, question=question, pattern=domain_pattern, remote_ip=remote_ip, remote_port=remote_port)
                return

        # answer question with default pattern
        default_pattern = None
        if self._answer_queries:
            for domain_pattern in self._configuration.get_domain_patterns():
                if domain_pattern.does_pattern_match(question=question):
                    default_pattern = domain_pattern

        # generate a new pattern and add it to the config file
        new_baseline_pattern = self._configuration.add_new_baseline_pattern(question=question, default_pattern=default_pattern)
        self._number_of_added_patterns += 1
        info('added new pattern to baseline:  {}  -  {}'.format(question.rr_type.name.ljust(4, ' '), question.name))

        self._answer_query_with_pattern(message=message, question=question, pattern=new_baseline_pattern, remote_ip=remote_ip, remote_port=remote_port)

    def get_number_of_added_patterns(self) -> int:
        return self._number_of_added_patterns
